/*********************************************************************************************************************************
Maximite

File: DrawChar.c

Implement a graphic font and associated functions for writing to a video device

Copyright 2011 Geoff Graham - http://geoffg.net
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General
Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.  You should have received a copy of the GNU General Public License along with this program.
If not, see <http://www.gnu.org/licenses/>. 

*********************************************************************************************************************************/

#include <p32xxxx.h>												// device specific defines
#include <plib.h>													// peripheral libraries

#include "../Maximite.h"											// helpful defines for the PIC32
#include "Video.h"


// Dimensions of one char in this font is 5 bits wide and 10 bits high
// Even though this is a 5x10 font we need to leave two blank horizontal rows of pixels below the char.
// We also need to leave one vertical column of blank pixels after the char.  
// This is why the vertical size is 12 and the horizontal is 6.
// if you change VSIZE you chould also change the define in Video.h
#define font_graphic_VSIZE	12
#define font_graphic_HSIZE	6

#define first_char 32  // First Character with Font Defined
#define last_char 191  // Last Character with Font Defined

const uint16 TEXT_graphic[160][6] = { 
					     0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  // SPACE - ASCII (32)
                         0x00, 0x00, 0x5F, 0x00, 0x00, 0x00,  // ! - ASCII (33) 
                         0x00, 0x03, 0x00, 0x03, 0x00, 0x00,  // " - ASCII (34)
                         0x14, 0x3E, 0x14, 0x3E, 0x14, 0x00,  // # - ASCII (35)
                         0x24, 0x2A, 0x7F, 0x2A, 0x12, 0x00,  // $ - ASCII (36)
                         0x43, 0x33, 0x08, 0x66, 0x61, 0x00,  // % - ASCII (37)
                         0x36, 0x49, 0x55, 0x22, 0x50, 0x00,  // & - ASCII (38)
                         0x00, 0x05, 0x03, 0x00, 0x00, 0x00,  // ' - ASCII (39)
                         0x00, 0x1C, 0x22, 0x41, 0x00, 0x00,  // ( - ASCII (40)
                         0x00, 0x41, 0x22, 0x1C, 0x00, 0x00,  // ) - ASCII (41)
                         0x14, 0x08, 0x3E, 0x08, 0x14, 0x00,  // * - ASCII (42)
                         0x08, 0x08, 0x3E, 0x08, 0x08, 0x00,  // + - ASCII (43)
                         0x00, 0x50, 0x30, 0x00, 0x00, 0x00,  // , - ASCII (44)
                         0x08, 0x08, 0x08, 0x08, 0x08, 0x00,  // - - ASCII (45)
                         0x00, 0x60, 0x60, 0x00, 0x00, 0x00,  // . - ASCII (46)
                         0x20, 0x10, 0x08, 0x04, 0x02, 0x00,  // / - ASCII (47)
                         0x3E, 0x51, 0x49, 0x45, 0x3E, 0x00,  // 0 - ASCII (48)
                         0x04, 0x02, 0x7F, 0x00, 0x00, 0x00,  // 1 - ASCII (49)
                         0x42, 0x61, 0x51, 0x49, 0x46, 0x00,  // 2 - ASCII (50)
                         0x22, 0x41, 0x49, 0x49, 0x36, 0x00,  // 3 - ASCII (51)
                         0x18, 0x14, 0x12, 0x7F, 0x10, 0x00,  // 4 - ASCII (52)
                         0x27, 0x45, 0x45, 0x45, 0x39, 0x00,  // 5 - ASCII (53)
                         0x3E, 0x49, 0x49, 0x49, 0x32, 0x00,  // 6 - ASCII (54)
                         0x01, 0x01, 0x71, 0x09, 0x07, 0x00,  // 7 - ASCII (55)
                         0x36, 0x49, 0x49, 0x49, 0x36, 0x00,  // 8 - ASCII (56)
                         0x26, 0x49, 0x49, 0x49, 0x3E, 0x00,  // 9 - ASCII (57)
                         0x00, 0x36, 0x36, 0x00, 0x00, 0x00,  // : - ASCII (58)
                         0x00, 0x56, 0x36, 0x00, 0x00, 0x00,  // ; - ASCII (59)
                         0x08, 0x14, 0x22, 0x41, 0x00, 0x00,  // < - ASCII (60)
                         0x14, 0x14, 0x14, 0x14, 0x14, 0x00,  // = - ASCII (61)
                         0x00, 0x41, 0x22, 0x14, 0x08, 0x00,  // > - ASCII (62)
                         0x02, 0x01, 0x51, 0x09, 0x06, 0x00,  // ? - ASCII (63)
                         0x3E, 0x41, 0x59, 0x55, 0x5E, 0x00,  // @ - ASCII (64)
                         0x7E, 0x09, 0x09, 0x09, 0x7E, 0x00,  // A - ASCII (65)
                         0x7F, 0x49, 0x49, 0x49, 0x36, 0x00,  // B - ASCII (66)
                         0x3E, 0x41, 0x41, 0x41, 0x22, 0x00,  // C - ASCII (67)
                         0x7F, 0x41, 0x41, 0x41, 0x3E, 0x00,  // D - ASCII (68)
                         0x7F, 0x49, 0x49, 0x49, 0x41, 0x00,  // E - ASCII (69)
                         0x7F, 0x09, 0x09, 0x09, 0x01, 0x00,  // F - ASCII (70)
                         0x3E, 0x41, 0x41, 0x49, 0x3A, 0x00,  // G - ASCII (71)
                         0x7F, 0x08, 0x08, 0x08, 0x7F, 0x00,  // H - ASCII (72)
                         0x00, 0x41, 0x7F, 0x41, 0x00, 0x00,  // I - ASCII (73)
                         0x30, 0x40, 0x40, 0x40, 0x3F, 0x00,  // J - ASCII (74)
                         0x7F, 0x08, 0x14, 0x22, 0x41, 0x00,  // K - ASCII (75)
                         0x7F, 0x40, 0x40, 0x40, 0x40, 0x00,  // L - ASCII (76)
                         0x7F, 0x02, 0x0C, 0x02, 0x7F, 0x00,  // M - ASCII (77)
                         0x7F, 0x02, 0x04, 0x08, 0x7F, 0x00,  // N - ASCII (78)
                         0x3E, 0x41, 0x41, 0x41, 0x3E, 0x00,  // O - ASCII (79)
                         0x7F, 0x09, 0x09, 0x09, 0x06, 0x00,  // P - ASCII (80)
                         0x1E, 0x21, 0x21, 0x21, 0x5E, 0x00,  // Q - ASCII (81)
                         0x7F, 0x09, 0x09, 0x09, 0x76, 0x00,  // R - ASCII (82)
						 0x26, 0x49, 0x49, 0x49, 0x32, 0x00,  // S - ASCII (83)
                         0x01, 0x01, 0x7F, 0x01, 0x01, 0x00,  // T - ASCII (84)
                         0x3F, 0x40, 0x40, 0x40, 0x3F, 0x00,  // U - ASCII (85)
                         0x1F, 0x20, 0x40, 0x20, 0x1F, 0x00,  // V - ASCII (86)
                         0x7F, 0x20, 0x10, 0x20, 0x7F, 0x00,  // W - ASCII (87)
                         0x41, 0x22, 0x1C, 0x22, 0x41, 0x00,  // X - ASCII (88)
                         0x07, 0x08, 0x70, 0x08, 0x07, 0x00,  // Y - ASCII (89)
                         0x61, 0x51, 0x49, 0x45, 0x43, 0x00,  // Z - ASCII (90)
                         0x00, 0x7F, 0x41, 0x00, 0x00, 0x00,  // [ - ASCII (91)
                         0x02, 0x04, 0x08, 0x10, 0x20, 0x00,  // backslash - ASCII (92)
                         0x00, 0x00, 0x41, 0x7F, 0x00, 0x00,  // ] - ASCII (93)
                         0x04, 0x02, 0x01, 0x02, 0x04, 0x00,  // ^ - ASCII (94)
                         0x40, 0x40, 0x40, 0x40, 0x40, 0x00,  // _ - ASCII (95)
                         0x00, 0x01, 0x02, 0x04, 0x00, 0x00,  // ` - ASCII (96)
                         0x20, 0x54, 0x54, 0x54, 0x78, 0x00,  // a - ASCII (97)
                         0x7F, 0x44, 0x44, 0x44, 0x38, 0x00,  // b - ASCII (98)
                         0x38, 0x44, 0x44, 0x44, 0x44, 0x00,  // c - ASCII (99)
                         0x38, 0x44, 0x44, 0x44, 0x7F, 0x00,  // d - ASCII (100)
                         0x38, 0x54, 0x54, 0x54, 0x18, 0x00,  // e - ASCII (101)
                         0x04, 0x04, 0x7E, 0x05, 0x05, 0x00,  // f - ASCII (102)
                         56, 580, 580, 580, 504, 0, 		  // g - ASCII (103)
                         0x7F, 0x08, 0x04, 0x04, 0x78, 0x00,  // h - ASCII (104)
                         0x00, 0x44, 0x7D, 0x40, 0x00, 0x00,  // i - ASCII (105)
                         256, 512, 516, 509, 0, 0,            // j - ASCII (106)
                         0x7F, 0x10, 0x28, 0x44, 0x00, 0x00,  // k - ASCII (107)
                         0x00, 0x41, 0x7F, 0x40, 0x00, 0x00,  // l - ASCII (108)
                         0x7C, 0x04, 0x78, 0x04, 0x78, 0x00,  // m - ASCII (109)
                         0x7C, 0x08, 0x04, 0x04, 0x78, 0x00,  // n - ASCII (110)
                         0x38, 0x44, 0x44, 0x44, 0x38, 0x00,  // o - ASCII (111)
                         1016, 68, 68, 68, 56,0,              // p - ASCII (112)
                         56, 68, 68, 68, 1016,0,              // q - ASCII (113)
                         0x00, 0x7C, 0x08, 0x04, 0x04, 0x00,  // r - ASCII (114)
                         0x48, 0x54, 0x54, 0x54, 0x20, 0x00,  // s - ASCII (115)
                         0x04, 0x04, 0x3F, 0x44, 0x44, 0x00,  // t - ASCII (116)
                         0x3C, 0x40, 0x40, 0x20, 0x7C, 0x00,  // u - ASCII (117)
                         0x1C, 0x20, 0x40, 0x20, 0x1C, 0x00,  // v - ASCII (118)
                         0x3C, 0x40, 0x30, 0x40, 0x3C, 0x00,  // w - ASCII (119)
                         0x44, 0x28, 0x10, 0x28, 0x44, 0x00,  // x - ASCII (120)
                         60, 576, 576, 576, 508,0,            // y - ASCII (121)
                         0x44, 0x64, 0x54, 0x4C, 0x44, 0x00,  // z - ASCII (122)
                         0x00, 0x08, 0x36, 0x41, 0x41, 0x00,  // { - ASCII (123)
                         0x00, 0x00, 0x7F, 0x00, 0x00, 0x00,  // | - ASCII (124)
                         0x41, 0x41, 0x36, 0x08, 0x00, 0x00,  // } - ASCII (125)
                         0x02, 0x01, 0x02, 0x04, 0x02, 0x00,  // ~ - ASCII (126)
                         0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x00,  // cursor (127)

            			// Graphics Set from System 80 / TRS-80

						0x000, 0x000, 0x000, 0x000, 0x000, 0x000,   // ASCII (128) - 0
						0x00f, 0x00f, 0x00f, 0x000, 0x000, 0x000,   // ASCII (129) - 1
						0x000, 0x000, 0x000, 0x00f, 0x00f, 0x00f,   // ASCII (130) - 2
						0x00f, 0x00f, 0x00f, 0x00f, 0x00f, 0x00f,   // ASCII (131) - 3
						0x0f0, 0x0f0, 0x0f0, 0x000, 0x000, 0x000,   // ASCII (132) - 4
						0x0ff, 0x0ff, 0x0ff, 0x000, 0x000, 0x000,   // ASCII (133) - 5
						0x0f0, 0x0f0, 0x0f0, 0x00f, 0x00f, 0x00f,   // ASCII (134) - 6
						0x0ff, 0x0ff, 0x0ff, 0x00f, 0x00f, 0x00f,   // ASCII (135) - 7
						0x000, 0x000, 0x000, 0x0f0, 0x0f0, 0x0f0,   // ASCII (136) - 8
						0x00f, 0x00f, 0x00f, 0x0f0, 0x0f0, 0x0f0,   // ASCII (137) - 9
						0x000, 0x000, 0x000, 0x0ff, 0x0ff, 0x0ff,   // ASCII (138) - 10
						0x00f, 0x00f, 0x00f, 0x0ff, 0x0ff, 0x0ff,   // ASCII (139) - 11
						0x0f0, 0x0f0, 0x0f0, 0x0f0, 0x0f0, 0x0f0,   // ASCII (140) - 12
						0x0ff, 0x0ff, 0x0ff, 0x0f0, 0x0f0, 0x0f0,   // ASCII (141) - 13
						0x0f0, 0x0f0, 0x0f0, 0x0ff, 0x0ff, 0x0ff,   // ASCII (142) - 14
						0x0ff, 0x0ff, 0x0ff, 0x0ff, 0x0ff, 0x0ff,   // ASCII (143) - 15

						0xf00, 0xf00, 0xf00, 0x000, 0x000, 0x000,   // ASCII (144) - 16
						0xf0f, 0xf0f, 0xf0f, 0x000, 0x000, 0x000,   // ASCII (145) - 17
						0xf00, 0xf00, 0xf00, 0x00f, 0x00f, 0x00f,   // ASCII (146) - 18
						0xf0f, 0xf0f, 0xf0f, 0x00f, 0x00f, 0x00f,   // ASCII (147) - 19
						0xff0, 0xff0, 0xff0, 0x000, 0x000, 0x000,   // ASCII (148) - 20
						0xfff, 0xfff, 0xfff, 0x000, 0x000, 0x000,   // ASCII (149) - 21
						0xff0, 0xff0, 0xff0, 0x00f, 0x00f, 0x00f,   // ASCII (150) - 22
						0xfff, 0xfff, 0xfff, 0x00f, 0x00f, 0x00f,   // ASCII (151) - 23
						0xf00, 0xf00, 0xf00, 0x0f0, 0x0f0, 0x0f0,   // ASCII (152) - 24
						0xf0f, 0xf0f, 0xf0f, 0x0f0, 0x0f0, 0x0f0,   // ASCII (153) - 25
						0xf00, 0xf00, 0xf00, 0x0ff, 0x0ff, 0x0ff,   // ASCII (154) - 26
						0xf0f, 0xf0f, 0xf0f, 0x0ff, 0x0ff, 0x0ff,   // ASCII (155) - 27
						0xff0, 0xff0, 0xff0, 0x0f0, 0x0f0, 0x0f0,   // ASCII (156) - 28
						0xfff, 0xfff, 0xfff, 0x0f0, 0x0f0, 0x0f0,   // ASCII (157) - 29
						0xff0, 0xff0, 0xff0, 0x0ff, 0x0ff, 0x0ff,   // ASCII (158) - 30
						0xfff, 0xfff, 0xfff, 0x0ff, 0x0ff, 0x0ff,   // ASCII (159) - 31

						0x000, 0x000, 0x000, 0xf00, 0xf00, 0xf00,   // ASCII (160) - 32
						0x00f, 0x00f, 0x00f, 0xf00, 0xf00, 0xf00,   // ASCII (161) - 33
						0x000, 0x000, 0x000, 0xf0f, 0xf0f, 0xf0f,   // ASCII (162) - 34
						0x00f, 0x00f, 0x00f, 0xf0f, 0xf0f, 0xf0f,   // ASCII (163) - 35
						0x0f0, 0x0f0, 0x0f0, 0xf00, 0xf00, 0xf00,   // ASCII (164) - 36
						0x0ff, 0x0ff, 0x0ff, 0xf00, 0xf00, 0xf00,   // ASCII (165) - 37
						0x0f0, 0x0f0, 0x0f0, 0xf0f, 0xf0f, 0xf0f,   // ASCII (166) - 38
						0x0ff, 0x0ff, 0x0ff, 0xf0f, 0xf0f, 0xf0f,   // ASCII (167) - 39
						0x000, 0x000, 0x000, 0xff0, 0xff0, 0xff0,   // ASCII (168) - 40
						0x00f, 0x00f, 0x00f, 0xff0, 0xff0, 0xff0,   // ASCII (169) - 41
						0x000, 0x000, 0x000, 0xfff, 0xfff, 0xfff,   // ASCII (170) - 42
						0x00f, 0x00f, 0x00f, 0xfff, 0xfff, 0xfff,   // ASCII (171) - 43
						0x0f0, 0x0f0, 0x0f0, 0xff0, 0xff0, 0xff0,   // ASCII (172) - 44
						0x0ff, 0x0ff, 0x0ff, 0xff0, 0xff0, 0xff0,   // ASCII (173) - 45
						0x0f0, 0x0f0, 0x0f0, 0xfff, 0xfff, 0xfff,   // ASCII (174) - 46
						0x0ff, 0x0ff, 0x0ff, 0xfff, 0xfff, 0xfff,   // ASCII (175) - 47

						0xf00, 0xf00, 0xf00, 0xf00, 0xf00, 0xf00,   // ASCII (176) - 48
						0xf0f, 0xf0f, 0xf0f, 0xf00, 0xf00, 0xf00,   // ASCII (177) - 49
						0xf00, 0xf00, 0xf00, 0xf0f, 0xf0f, 0xf0f,   // ASCII (178) - 50
						0xf0f, 0xf0f, 0xf0f, 0xf0f, 0xf0f, 0xf0f,   // ASCII (179) - 51
						0xff0, 0xff0, 0xff0, 0xf00, 0xf00, 0xf00,   // ASCII (180) - 52
						0xfff, 0xfff, 0xfff, 0xf00, 0xf00, 0xf00,   // ASCII (181) - 53
						0xff0, 0xff0, 0xff0, 0xf0f, 0xf0f, 0xf0f,   // ASCII (182) - 54
						0xfff, 0xfff, 0xfff, 0xf0f, 0xf0f, 0xf0f,   // ASCII (183) - 55
						0xf00, 0xf00, 0xf00, 0xff0, 0xff0, 0xff0,   // ASCII (184) - 56
						0xf0f, 0xf0f, 0xf0f, 0xff0, 0xff0, 0xff0,   // ASCII (185) - 57
						0xf00, 0xf00, 0xf00, 0xfff, 0xfff, 0xfff,   // ASCII (186) - 58
						0xf0f, 0xf0f, 0xf0f, 0xfff, 0xfff, 0xfff,   // ASCII (187) - 59
						0xff0, 0xff0, 0xff0, 0xff0, 0xff0, 0xff0,   // ASCII (188) - 60
						0xfff, 0xfff, 0xfff, 0xff0, 0xff0, 0xff0,   // ASCII (189) - 61
						0xff0, 0xff0, 0xff0, 0xfff, 0xfff, 0xfff,   // ASCII (190) - 62
						0xfff, 0xfff, 0xfff, 0xfff, 0xfff, 0xfff    // ASCII (191) - 63
};




// Write a char on the  video using the graphic bit font
void DrawChar(int x, int y, char c) {								// x and y are in pixels
    int j, k;                     									// Loop counters, j for horizontal, y for vertical
	
	for(j = 0; j < font_graphic_HSIZE; ++j)         				// Loop through each word (ie, horizontal scan line)
		for(k = 0; k < font_graphic_VSIZE; ++k)         			// Loop through the bits in the word (ie, vertical pixels)
			plot(x + j, y + k, (TEXT_graphic[c - first_char][j] & (1<< k)) != 0);	// Draw the pixel

	//for(k = 0; k < font_graphic_VSIZE; ++k) plot(x + j, y + k, 0);	// Draw a blank column of vertical pixels after the character
}

// Draw a Cursor as below the normal character positions.
void DrawCursor(int x, int y, char b) {
    int j, k;                     									// Loop counters, j for horizontal, y for vertical
	for(j = 0; j < font_graphic_HSIZE-1; ++j)         				// Loop through each word (ie, horizontal scan line)
		for(k = font_graphic_VSIZE-3; k < font_graphic_VSIZE; ++k)  // Loop through the bits in the word (ie, vertical pixels)
			plot(x + j, y + k, b);									// Invert the pixels at the Cursor Location - plot(x + j, y + k, (pixel(x+j,y+k) == 0));
}

/*********************************************************************************************************************************
 The following variables and functions are the only interface to the MMBasic engine
 They are defined in Video.h as extern
 *********************************************************************************************************************************/

// these keep track of where the next char will be written
int MMPosX, MMPosY;													// position in pixels
int MMCharPos;														// position in lines and characters


// Put a char onto the video screen
void VideoPutc(char c) {
	if(c == '\r') MMPosX = 0; 										// for a return set the horizontal position to zero
	if(c == '\n') MMPosY += font_graphic_VSIZE; 					// for a line feed add the char height to the vert position
	if(c == '\t')  													// for a tab move the horiz position to the next tab spot
		MMPosX = ((MMPosX + (font_graphic_HSIZE << 3)) >> 3) << 3;
	if(c == '\b') {													// for a backspace move the horiz position back by a char width
		MMPosX -= font_graphic_HSIZE;
		if(MMPosX <0) MMPosX = 0;
	}
	
	if(c >= first_char && c <= last_char) {							// printable chars only
		if(MMPosX + font_graphic_HSIZE > HRes) {					// if we are at the end of a line
			MMPosX = 0;												// wrap to the next
			MMPosY += font_graphic_VSIZE;
		}
	}	
		
	if(MMPosY + font_graphic_VSIZE > VRes + 1) {					// if we are beyond the bottom of the screen scroll up the previous lines
		char *pd = VA;
       	char *ps = pd + (HBUFSIZE/8)*font_graphic_VSIZE;
       	int i;
       	for( i=0; i<(HBUFSIZE/8) * (VRes-font_graphic_VSIZE); i++) *pd++ = *ps++;	// scroll up
       	for( i=0; i<(HBUFSIZE/8) * font_graphic_VSIZE; i++) *pd++ = 0;	// clear the new line
		MMPosY -= font_graphic_VSIZE;
	}
		
	if(c >= first_char && c <= last_char) {							// printable chars only
		DrawChar(MMPosX, MMPosY, c);								// draw the char
		MMPosX += font_graphic_HSIZE;								// and update our position
	}
	
	MMCharPos = MMPosX/font_graphic_HSIZE + 1;						// update the horizontal character position
}		



// draw/erase the cursor on the video screen
// if b is true draw the cursor else erase it
void MMCursor(char b) {
	DrawCursor(MMPosX, MMPosY,b);
	//
	//if(b)
    //	DrawChar(MMPosX, MMPosY, '~' + 1);
	//else
	//	DrawChar(MMPosX, MMPosY, ' ');
}	


